
# $Id: marker-rank.r 72 2016-03-10 14:31:06Z manidr $
print ("$Id: marker-rank.r 72 2016-03-10 14:31:06Z manidr $")



# library (RankAggreg)
# aggregate.ranks.optimal <- function (data, ranks.cols, id.col='id', sort.by.col="q.value", max.zeros=2, 
#                                      method='GA',          # options: CE (cross-entropy) or GA (genetic algo)
#                                      distance='Spearman',  # options: Spearman or Kendall
#                                      ...                   # other options to RankAggreg
# )
# {
#   ## Reference: Pihur, V., Datta, S., & Datta, S. (2009). 
#   #             RankAggreg, an R package for weighted rank aggregation. 
#   #             BMC Bioinformatics, 10, 62. http://doi.org/10.1186/1471-2105-10-62
#   #  Using RankAggreg is computationally intensive for cross-entropy MC (too slow, crashes), 
#   #  and does not converge for genetic algorithms (even with 10,000 iterations).
#   #  Seems to work only for small lists (10-50)
#   
#   ## function for aggregating multiple ranks for a gene/protein (i.e., each id) into a final ranking
#   ## the data is assumed to have importance scores (high => more important)
#   ## these will be converted to ranks (low => more important)
#   #
#   # rank aggregation is performed after the following filtering:
#   #  . eliminate ids with too many zeros (i.e., if id has >= max.zeros)
#   #  . sort the ids by increasing q-value (sort.by.col)
#   # this is needed since the RankAggreg function cannot deal with duplicate ranks,
#   # and unimportant ids should have high ranks
#   
#   d <- data [ order (data[,sort.by.col]), ]     # sort by q-value
#   d <- d [, c (id.col, ranks.cols)]             # subset for rank aggreration
#   no.id <- setdiff (colnames (d), id.col)
#   d <- d [ apply (d[,no.id], 1, function (x) sum (x==0)) < max.zeros, ]   # remove rows with too many 0's
#   
#   d.ranks <- apply (d[,no.id], 2, function (x) rank ( max(x) - x, ties='first'))
#   final.ranks <- RankAggreg (t(d.ranks), k=nrow(d.ranks), method=method, distance=distance, ...)
#   d <- d[final.ranks$top.list,]                 # reorder data by final rank
#   d.final <- cbind (d[,id.col], final.rank=1:nrow(d))    # record ranks
#   colnames (d.final) <- c (id.col, 'final.rank')
#   data.final <- merge (data, d.final, by=id.col, all.x=TRUE)
#   
#   return (data.final)
# } 


library (RobustRankAggreg)
aggregate.ranks <- function (data, ranks.cols, id.col='id', sort.by.col="q.value", 
                                     method='stuart',      # options: stuart, RRA
                                     ...                   # other options
)
{
  ## Reference: Kolde, R., Laur, S., Adler, P., & Vilo, J. (2012). 
  ##            Robust rank aggregation for gene list integration and meta-analysis. 
  ##            Bioinformatics, 28(4), 573–580. http://doi.org/10.1093/bioinformatics/btr709
  
  ## function for aggregating multiple ranks for a gene/protein (i.e., each id) into a final ranking
  ## the data is assumed to have importance scores (high => more important)
  ## these will be converted to ranks (low => more important)
  #
  # rank aggregation is performed after the following filtering:
  #  . sort the ids by increasing q-value (sort.by.col)
  # this is needed so that unimportant ids have high ranks
  
  d <- data [ order (data[,sort.by.col]), ]     # sort by q-value
  d <- d [, c (id.col, ranks.cols)]             # subset for rank aggreration
  no.id <- setdiff (colnames (d), id.col)
  
  d.ranks <- apply (d[,no.id], 2, function (x) rank (max(x) - x))
  rownames (d.ranks) <- d[,id.col]
  # use method='stuart' since this provides a score that we can sort on
  # using method='RRA' results in corrected p-values that are all 1
  final.ranks <- aggregateRanks (rmat=d.ranks, method='stuart', ...)  
  d.final <- cbind (final.ranks, final.rank=1:nrow(d))    # record ranks
  colnames (d.final) <- c (id.col, 'score', 'final.rank')
  data.final <- merge (data, d.final, by=id.col, all.x=TRUE)
  data.final <- data.final [ order (data.final[,'final.rank']), ]    # reorder by final rank
  
  return (data.final)
} 

