
# $Id: markersel-classify.r 73 2016-03-16 14:48:05Z manidr $
print ("$Id: markersel-classify.r 73 2016-03-16 14:48:05Z manidr $")


##
## AUTOMATED MARKER SELECTION AND CLASSIFICATION
#  Usage:
#  marker.selection.and.classification (
#    gct.file,                # Input file in gct format with columns representing samples
#                               and rows representing proteins/genes/phosphopeptides/etc.
#                               May contain missing values (NA)
#    cls.file,                # The class vector providing class labels for samples in the gct.file
#                               (order must match with sample order in gct.file)
#    output.prefix,           # Name prepended to all output file names
#    test.file=NULL,          # Test data (gct format) -- each classifier is used to predict class 
#                               membership for these samples; must contain the same proteins/genes/etc.
#                               present in the input gct.file
#    test.cls=NULL,           # Class vector for test samples
#    id.col=1,                # Column in gct file (either 1 or 2) that contains protein/gene ID
#    desc.col=2,              # Column in gct file (either 1 or 2) that contains protein/gene description
#    fdr=0.01,                # FDR p-value threshold for determining significant markers;
#                               Significant markers are used to train classifiers
#    log.tr=TRUE,             # Whether input data has been log transformed
#    cv.fold=10,              # Number of cross validation folds
#    models=c('pls', 'pam', 'rf', 'glmnet'),     # Classifiers to train -- 
#                                                  see documentation for caret library for other options
#    sample.weights=NULL      # Relative weights for samples to be used in classifier training
#    tune.steps=5             # Number of steps for each tuning parameter
#    heatmap=TRUE,            # Parameter for heatmaps -- if FALSE heatmap will not be generated
#    n.markers=0,             # Number of significant markers to show in heapmap -- show all if 0
#    heatmap.range=c(-3,3),   # Range of color scale for heatmap
#    heatmap.levels=12,       # Number of levels for color scale
#    heatmap.annotation.colors=NULL,           # colors to use for annotation bar; if NULL, use ColorBrewer Set1
#    original.data.for.heatmap=TRUE,           # If TRUE, heatmap shows original data (with NAs); 
#                                                else shows imputed data
#    heatmap.width=8, heatmap.height=11,       # Heatmap dimensions (in inches)
#    gsea=FALSE,                               # run GSEA for 2-class cls if TRUE
#    gsea.fdr=0.05, gsea.prefix='gsea',        # GSEA parameters; gsea.prefix is used in directory/file naming
#    gsea.pathway.db=file.path (Rutil.path, 'GSEA', 'msigdb.v4.0.symbols.gmt'),
#                             # location of the GSEA pathway database file (.gmt)
#    id.to.gene.map=file.path (Rutil.path, '..', 'gene-mapping', 'RefSeq.20130727-Human_gi-RefSeq-GeneName.txt'),
#                             # file for mapping IDs in gct.file to gene names
#                             # if gct.file already has gene names, set to NULL for no mapping
#    official.genenames=file.path (Rutil.path, '..', 'gene-mapping', 'std-gene-map', 'gene-symbol-map.csv'),
#                             # file containing official gene names with corresponding aliases 
#    map.id.col='refseq_protein',   # column name in id.to.gene.map file that has IDs in gct.file
#    map.genename.col='gene_name',  # column name in id.to.gene.map file that has gene names
#    duplicate.gene.policy='maxvar' # policy for combining duplicate genes (maxvar, median or union)
#  )
#
#  The first 3 arguments must be specified. The other arguments have defaults that will be used
#  if they are not specified.
#  
#  Output:
#  * Classifier performance contingency tables will be printed to output
#  * File <output-prefix>-markers-fdr<FDR>.csv
#       List of significant markers dervied using SAM. Includes protein/gene ID,
#       SAM score, fold change and q-value.
#       Fold change = mean in class 1 / mean in class 2.
#        (order of classes is determined by order of factor levels in the
#          class vector - usually alphabetical)
#  * File <output-prefix>-markersig-fdr<FDR>.csv
#       For multiclass (> 2 classes) cases, table showing classes for which a 
#       given marker is significant (along with direction: -1 or +1)
#  * File <output-prefix>-markerimp-fdr<FDR>.csv
#       Summary of marker importance for significant markers.
#  * File <output-prefix>-markers-heatmap.pdf
#       Heatmap for significant markers 
#  * File <output-prefix>-train-results.csv
#       Table of prediction results for training data from all classifiers
#  * File <output-prefix>-test-results.csv
#       Table of prediction results for test dataset (if provided) from all classifiers
#  * Directory <output-prefix>-<gsea.prefix>-analysis
#       Contains GSEA <gsea.prefix>-input.gct and <gsea.prefix>-input.cls files
#         which are used a GSEA inputs. The directory also has all GSEA output.
#         See GSEA documentation for details of output files/plots created.
#         Order of classes for GSEA is determined by the order in which they are
#          listed in the cls file (0=first listed class; 1=second listed class, etc.)
##
  


## for parsing GenePattern inputs
suppressMessages(suppressWarnings(library(getopt)))
suppressMessages(suppressWarnings(library(optparse)))


## libraries needed by this module
suppressMessages (suppressWarnings (library (caret)))
suppressMessages (suppressWarnings (library (samr)))
suppressMessages (suppressWarnings (library (pamr)))
suppressMessages (suppressWarnings (library (randomForest)))
suppressMessages (suppressWarnings (library (glmnet)))
suppressMessages (suppressWarnings (library (NMF)))
suppressMessages (suppressWarnings (library (tools)))
suppressMessages (suppressWarnings (library (RColorBrewer)))


sessionInfo ()
set.seed (2)


marker.selection.and.classification <- function (gct.file, cls.file, output.prefix, 
                                                 test.file=NULL, test.cls=NULL,
                                                 id.col=1, desc.col=2, impute.colmax=0.8,
                                                 output.imputed.data=FALSE,
                                                 fdr=0.01, log.tr=TRUE, cv.fold=10,
                                                 output.all.markers=TRUE,
                                                 models=c('pls', 'pam', 'rf', 'glmnet'), 
                                                 sample.weights=NULL, tune.steps=5,
                                                 heatmap=TRUE, n.markers=0, heatmap.range=c(-3,3),
                                                 heatmap.annotation.colors=NULL,
                                                 heatmap.levels=12, original.data.for.heatmap=TRUE,
                                                 heatmap.width=8, heatmap.height=11,
                                                 # GSEA and gene mapping parameters -- DISABLE
                                                 gsea=FALSE, gsea.fdr=0.05, gsea.prefix='gsea',
                                                 gsea.pathway.db=file.path (Rutil.path, 'GSEA', 'c2.all.v4.0.symbols-with-ischemia-sig.gmt'),
                                                 id.to.gene.map=file.path (Rutil.path, '..', 'gene-mapping', 'RefSeq.20130727-Human_gi-RefSeq-GeneName.txt'),
                                                 official.genenames=file.path (Rutil.path, '..', 'gene-mapping', 'std-gene-map', 'gene-symbol-map.csv'),
                                                 map.id.col='refseq_protein', map.genename.col='gene_name',
                                                 duplicate.gene.policy="maxvar") {
  ## runs marker selection using SAM
  ## use the best markers to build classfiers using PLS, PAM and RF
  ## output best markers and (cross validation) classifier results
  ## if data has missing values:
  ##  - filter to remove features with > 50% mising
  ##  - impute missing values using k-NN
  ## if test data is provided, run prediction using all classifiers
  ## plots heatmap of training data showing significant markers
  
  
  read.filter.impute <- function (gct.file, cls.file, nmiss.factor=0.5) {
    d <- read.gct (gct.file, check.names=FALSE)
    rownames (d) <- d [, id.col]
    data <- d [, c(-1,-2)]
    if (!is.null (cls.file)) {
      cls <- factor (read.cls (cls.file))
      if ( ncol(data) != length(cls) ) stop ('Class vector ', cls.file, ' does not match data in ', gct.file)
    } else {
      cls <- NULL
    }
    ## filter for missing values
    n.miss <- apply (data, 1, function (x) sum (is.na (x)))
    keep <- n.miss < (ncol(data) * nmiss.factor)
    d <- d [ keep, ]
    data <- d [, c(-1,-2)]
    
    ## impute missing values
    data.noNA <- pamr.knnimpute ( data=list (x=as.matrix(data), y=cls), colmax=impute.colmax )$x
    return (list (dataset=d, imputed.dataset=data.noNA, class.vector=cls))
  }
  
  
  ## training data
  training.data <- read.filter.impute (gct.file, cls.file)
  d <- training.data$dataset
  data.noNA <- training.data$imputed.dataset
  cls <- training.data$class.vector
  # write out imputed dataset
  if (output.imputed.data) {
    imputed.file <- paste (file_path_sans_ext (gct.file), '-imputed.gct', sep='')
    if (! file.exists (imputed.file) ) write.gct (cbind (d[,1:2], data.noNA), imputed.file)
    else warning (paste ("File", imputed.file, "already exists. NOT over-written"))
  }
  
  
  ## test data -- if any
  if (!is.null (test.file)) testing.data <- read.filter.impute (test.file, test.cls, nmiss.factor=1)
  

  ## marker selection function
  sam <- function (FDR) {
    markersel <- SAM (x=as.matrix (data.noNA), y=cls, genenames=as.character (d[,id.col]),
                      resp.type=ifelse (nlevels (cls) > 2, "Multiclass", "Two class unpaired"),
                      nperms=1000, fdr.output=FDR, logged2=log.tr)
    # output results
    markers <- NULL
    if (markersel$siggenes.table$ngenes.up > 0) markers <- rbind (markers, markersel$siggenes.table$genes.up)
    if (markersel$siggenes.table$ngenes.lo > 0) markers <- rbind (markers, markersel$siggenes.table$genes.lo)  
    return ( list (markers=markers, sam.results=markersel) )
  }
  
  
  ## calculate q-values for all input features and save results
  if (output.all.markers) {
    markers.all <- sam (FDR=1)
    write.csv (markers.all$markers, paste (output.prefix, '-markers-all.csv', sep=''), row.names=FALSE)
  }
  
  
  ## marker selection
  markers.fdr <- sam (fdr)
  markers <- markers.fdr$markers
  markersel <- markers.fdr$sam.results
  if (nrow (markers) < 1) stop ('No significant markers -- try increasing fdr threshold')
  write.csv (markers, paste (output.prefix, '-markers-fdr', fdr, '.csv', sep=''), row.names=FALSE)
  if (!is.null (markersel$siggenes.table$color.ind.for.multi)) 
    write.csv (markersel$siggenes.table$color.ind.for.multi, 
               paste (output.prefix, '-markersig-fdr', fdr, '.csv', sep=''), row.names=FALSE)

  
  ## create data subset for training/testing classfiers (using caret)
  create.subset <- function (d.all, d.nomiss) {
    subset <- d.all [, id.col] %in% markers[,'Gene ID']
    d.sig <- d.all [ subset, ]
    data.sig <- d.nomiss [ subset,]
    data.subset <- t (data.sig)
    return (data.subset)
  }
  data.train <- create.subset (d, data.noNA)
  train.pred <- cbind (sample=rownames (data.train), actual=as.character (cls))
  if (!is.null (test.file)) {
    data.test <- create.subset (testing.data$dataset, testing.data$imputed.dataset)
    test.pred <- cbind (sample=rownames (data.test))
    if (!is.null (test.cls)) test.pred <- cbind (test.pred, actual=as.character (testing.data$class.vector))
  }
  
  
  ## train / test models, evaluate performance and track variable importance
  model.predict <- function (m, model, new.data=NULL) {
    pred.factor <- predict (model, newdata=new.data)
    pred <- as.character (pred.factor)
    prob <- extractProb (list(model), unkX=new.data)
    prediction.prob <- sapply ( 1:length(pred), function (i) prob[i, make.names(pred[i])] )
    prediction.prob.diff <- sapply ( 1:length(pred), 
                                     function (i) {
                                       probs <- sort (prob[i, make.names (levels (pred.factor))], decreasing=TRUE)
                                       probs[1] - probs[2]  # difference between the highest and next best probabilities
                                     } )
    result <- cbind (pred, prediction.prob, prediction.prob.diff)
    colnames (result) <- c (m, paste (m, '.prob', sep=''), paste (m, '.prob.topdiff', sep=''))
    return (result)
  }
  varimp <- markers [, setdiff (colnames (markers), c("Gene Name", "Numerator(r)", "Denominator(s+s0)"))]
  colnames (varimp)[1] <- 'id'
  for (m in models) {
    model <- train (x=data.train, y=cls, method=m, weights=sample.weights, 
                    trControl=trainControl (method='cv', number=cv.fold), tuneLength=tune.steps)
    cat ('\n\n------------------------------------------------------------\n')
    print (model)
    cat ('\n\nCONFUSION MATRIX: Cross-Validation\n')
    print (confusionMatrix (model, norm='average'))
    # variable importance
    if (m=='glmnet') {
      # use (absolute) coefficients as variable importance for Elastic Nets
      enet.coef <- coef (model$finalModel, s=model$bestTune$lambda)
      if (nlevels (factor(cls)) == 2) {
        imp <- as.matrix (enet.coef)
      } else {
        imp <- Reduce (function (c1, c2) cbind (as.matrix (c1), as.matrix (c2)), enet.coef)
        colnames (imp) <- names (enet.coef)
      }
      vi <- data.frame (imp[-1,])    # remove intercept      
    } else {
      vi <- varImp (model)$importance
    }
    vi.table <- data.frame (rownames (vi), apply (vi, 1, function (x) abs (max (x))))
    colnames (vi.table) <- c ('id', m)
    varimp <- merge (varimp, vi.table, by='id')
    # prediction on train data
    pred.tr <- model.predict (m, model)
    train.pred <- cbind (train.pred, pred.tr)
    cat ('\nCONFUSION MATRIX: Training Data\n')
    print (table (predicted=unlist (pred.tr[,1]), actual=unlist (train.pred[,'actual'])))
    # prediction on test data
    if (!is.null (test.file)) {
      pred.te <- model.predict (m, model, new.data=data.test)
      test.pred <- cbind (test.pred, pred.te)
      if ('actual' %in% colnames (test.pred)) {
        cat ('\nCONFUSION MATRIX: Testing Data\n')
        print (table (predicted=unlist (pred.te[,1]), actual=unlist (test.pred[,'actual'])))
      }
    }
  }  
  
  
  ## process variable importance table
  samscore <- abs (as.numeric (as.character (varimp[,"Score(d)"])))
  varimp <- cbind (varimp, sam=samscore*100/max (samscore))
  varimp <- varimp [ order (samscore, decreasing=TRUE), ]
  # include protein descriptions
  d.info <- data.frame (id=d[,id.col], description=d[,desc.col])
  varimp.final <- merge (varimp, d.info, by='id')  
  # create combined final ranking for markers
  varimp.final <- aggregate.ranks (varimp.final, ranks.cols=models, id.col='id', sort.by.col='q-value(%)')
  write.csv (varimp.final, paste (output.prefix, '-markerimp-fdr', fdr, '.csv', sep=''), row.names=FALSE)  
  
  ## write out train and test data predictions
  write.csv (train.pred, paste (output.prefix, '-train-results-fdr', fdr, '.csv', sep=''), row.names=FALSE)
  if (!is.null (test.file)) 
    write.csv (test.pred, paste (output.prefix, '-test-results-fdr', fdr, '.csv', sep=''), row.names=FALSE)
  
  
  ## heatmap
  if (heatmap) {
    # reorder training data by class
    cls.sorted <- order (cls)
    annotation <- data.frame (class=cls [cls.sorted])
    n.classes <- max (apply (annotation, 2, function (x) nlevels (factor(x))))
    if (original.data.for.heatmap) {
      subset <- d [, id.col] %in% markers[,'Gene ID']
      plot.data <- d [ subset, c(-1,-2)]      
    } else plot.data <- t (data.train)
    plot.data <- plot.data[, cls.sorted]   # resort by class
    if (n.markers > 0) {
      # select best marker subset
      id.sort <- order ( abs (varimp.final[,'Score(d']), decreasing=TRUE )
      is.subset <- id.sort [1:n.markers]
      plot.data <- plot.data [ id.subset, ]
    }
    # create heatmap
    # aheatmap makes it easy to annotate heatmap; other options are heatmap.2 and heatmap.3
    nmf.options (grid.patch=TRUE)  # to avoid blank page in PDF output
    pdf ( paste (output.prefix, '-markers-heatmap-fdr', fdr, '.pdf', sep=''), width=heatmap.width, height=heatmap.height)
    if (is.null (heatmap.annotation.colors) || length (heatmap.annotation.colors) != n.classes ) 
      heatmap.annotation.colors <- brewer.pal (n.classes, "Set1")
    aheatmap (plot.data, Colv=NA, annCol=annotation, annColors=list (heatmap.annotation.colors),
              scale='row', color='-RdBu', 
              breaks=seq (heatmap.range[1], heatmap.range[2], length.out=heatmap.levels))
    dev.off()
  }
  
  
  ## GSEA
  # Run GSEA (on training data) if class vector is two-class
  if (nlevels (factor (cls)) == 2 && gsea) {
    Source ('GSEA/GSEA.1.0.R')
    
    # sub-directory for GSEA results
    results.dir <- paste (output.prefix, '-gsea-analysis', sep='')
    if (file.exists (results.dir)) {
      # directory exists -- create a variant name (using a brute force approach!)
      i <- 0
      unique.results.dir <- results.dir
      while (file.exists (unique.results.dir)) {
        i <- i+1
        unique.results.dir <- paste (results.dir, '_', i, sep='')
      }
      results.dir <- unique.results.dir
    }
    dir.create (results.dir)
        
    if (!is.null (id.to.gene.map)) {
      # map protein IDs in id.col to gene names (if requested)
      # use NA-filtered + imputed dataset
      d.gsea <- map.proteins.to.genes (cbind (d.info, data.noNA), id.col=1, id.to.gene.map,
                                       protein.id.col=map.id.col, gene.id.col=map.genename.col,
                                       keep.cols=c(map.id.col, colnames (data.noNA)))
      if (nrow (d.gsea) == 0)
        # gene mapping failed -- try using a different id.col
        # (in some cases, the protein id is the second column)
        d.gsea <- map.proteins.to.genes (cbind (d.info, data.noNA), id.col=2, id.to.gene.map,
                                         protein.id.col=map.id.col, gene.id.col=map.genename.col,
                                         keep.cols=c(map.id.col, colnames (data.noNA)))
      if (nrow (d.gsea) == 0)
        # nothing worked -- give up
        stop ('Gene mapping failed ... giving up.')
    } else {
      d.gsea <- cbind (d.info, data.noNA)
    }
    gsea.input <- process.duplicate.genes (d.gsea, genesym.col=1, data.cols=3:ncol(d.gsea), 
                                           official.symbols.file=official.genenames, map.genes=TRUE,
                                           policy=duplicate.gene.policy) 

    # write out gct and cls file -- direct input of data/cls results in incompatibilities with GSEA expectations
    gsea.gct <- file.path (results.dir, paste (gsea.prefix, '-input.gct', sep=''))
    gsea.cls <- file.path (results.dir, paste (gsea.prefix, '-input.cls', sep=''))
    write.gct (gsea.input, gsea.gct)
    write.cls (cls, gsea.cls)
    
    # run GSEA on all MSigDB pathways
    gsea.result <- GSEA (gsea.gct, gsea.cls, gs.db=gsea.pathway.db, output.directory=paste (results.dir,'/',sep=''), 
                         doc.string=gsea.prefix, fdr.q.val.threshold=gsea.fdr)
  }
}




# Get the command line arguments.  We'll process these with optparse.
# https://cran.r-project.org/web/packages/optparse/index.html
arguments <- commandArgs(trailingOnly=TRUE)

# Declare an option list for optparse to use in parsing the command line.
option_list <- list(
  # Note: it's not necessary for the names to match here, it's just a convention
  # to keep things consistent.
  make_option("--libdir", dest="libdir"),
  make_option("--gct.file", dest="input.file"),
  make_option("--gct.id.col", dest="id.col", type='integer', default=2),
  make_option("--cls.file", dest="cls.file"),
  make_option("--output.prefix", dest="output.prefix"),
  make_option("--fdr.threshold", dest="fdr", type='numeric', default=0.01)
)

# Parse the command line arguments with the option list, printing the result
# to give a record as with sessionInfo.
opt <- parse_args (OptionParser (option_list=option_list), positional_arguments=TRUE, args=arguments)
print (opt)
opts <- opt$options


## read in supporting functions
source (file.path (opts$libdir, "io.r"))
source (file.path (opts$libdir, "marker-rank.r"))


## Run marker selection/ranking and corresponding classifiers
with (opts,
      marker.selection.and.classification (input.file, cls.file, output.prefix, 
                                           fdr=fdr,
                                           id.col=id.col, desc.col=ifelse (id.col==1, 2, 1))
)

